from typing import Iterator, List, Dict, Optional
import requests

from aistore.sdk.request_client import RequestClient
from aistore.sdk.const import DEFAULT_CHUNK_SIZE, HTTP_METHOD_GET, HTTP_METHOD_HEAD
from aistore.sdk.object_attributes import ObjectAttributes


class ObjectReader:
    """
    Represents the data returned by the API when getting an object, including access to the content stream and object
    attributes.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        client: RequestClient,
        path: str,
        params: List[str],
        headers: Optional[Dict[str, str]] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ):
        self._request_client = client
        self._request_path = path
        self._request_params = params
        self._request_headers = headers
        self._chunk_size = chunk_size
        self._attributes = None

    def _head(self) -> ObjectAttributes:
        """
        Make a head request to AIS to update and return only object attributes.

        Returns:
            ObjectAttributes for this object

        """
        resp = self._request_client.request(
            HTTP_METHOD_HEAD, path=self._request_path, params=self._request_params
        )
        return ObjectAttributes(resp.headers)

    def _make_request(self, stream) -> requests.Response:
        """
        Make a request to AIS to get the object content.

        Returns:
            requests.Response from AIS

        """
        resp = self._request_client.request(
            HTTP_METHOD_GET,
            path=self._request_path,
            params=self._request_params,
            stream=stream,
            headers=self._request_headers,
        )
        self._attributes = ObjectAttributes(resp.headers)
        return resp

    @property
    def attributes(self) -> ObjectAttributes:
        """
        Object metadata attributes.

        Returns:
            ObjectAttributes: Parsed object attributes from the headers returned by AIS
        """
        if not self._attributes:
            self._attributes = self._head()
        return self._attributes

    def read_all(self) -> bytes:
        """
        Read all byte data directly from the object response without using a stream.

        This requires all object content to fit in memory at once and downloads all content before returning.

        Returns:
            bytes: Object content as bytes.
        """
        return self._make_request(stream=False).content

    def raw(self) -> requests.Response:
        """
        Returns the raw byte stream of object content.

        Returns:
            requests.Response: Raw byte stream of the object content
        """
        return self._make_request(stream=True).raw

    def __iter__(self) -> Iterator[bytes]:
        """
        Make a request to get a stream from the provided object and yield chunks of the stream content.

        Returns:
            Iterator[bytes]: An iterator over each chunk of bytes in the object
        """
        stream = self._make_request(stream=True)
        try:
            yield from stream.iter_content(chunk_size=self._chunk_size)
        finally:
            stream.close()



#
# Copyright (c) 2022-2024, NVIDIA CORPORATION. All rights reserved.
#
from urllib.parse import urljoin, urlencode
from typing import Callable, TypeVar, Type, Any, Dict, Optional, Tuple, Union
from requests import Response

from aistore.sdk.const import (
    JSON_CONTENT_TYPE,
    HEADER_USER_AGENT,
    USER_AGENT_BASE,
    HEADER_CONTENT_TYPE,
    HEADER_AUTHORIZATION,
)
from aistore.sdk.session_manager import SessionManager
from aistore.sdk.utils import raise_ais_error, handle_errors, decode_response
from aistore.version import __version__ as sdk_version

T = TypeVar("T")


# pylint: disable=unused-variable, duplicate-code, too-many-arguments
class RequestClient:
    """
    Internal client for making requests to an AIS cluster or a specific proxy (e.g. AuthN).

    Args:
        endpoint (str): Endpoint to either the AIS cluster or a specific proxy (e.g. AuthN).
        session_manager (SessionManager): SessionManager for creating and accessing requests session.
        timeout (Union[float, Tuple[float, float], None], optional): Request timeout in seconds; a single float.
            for both connect/read timeouts (e.g., 5.0), a tuple for separate connect/read timeouts (e.g., (3.0, 10.0)),
            or None to disable timeout.
        token (str, optional): Authorization token.
        error_handler (Callable[[str], None], optional): Error handler for managing response errors. Defaults to
            raise_ais_error.
    """

    def __init__(
        self,
        endpoint: str,
        session_manager: SessionManager,
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        token: str = None,
        error_handler: Callable[[str], None] = raise_ais_error,
    ):
        self._base_url = urljoin(endpoint, "v1")
        self._session_manager = session_manager
        self._token = token
        self._timeout = timeout
        self._error_handler = error_handler

    @property
    def base_url(self):
        """Return the base URL."""
        return self._base_url

    @property
    def timeout(self):
        """Return the timeout for requests."""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: Union[float, Tuple[float, float]]):
        """
        Set the timeout for all requests from this client.

        Args:
            timeout: Timeout for requests.
        """
        self._timeout = timeout

    @property
    def session_manager(self) -> SessionManager:
        """
        Return the SessionManager used to create sessions for this client.
        """
        return self._session_manager

    @session_manager.setter
    def session_manager(self, session_manager):
        self._session_manager = session_manager

    @property
    def token(self) -> str:
        """
        Return the token for authorization.
        """
        return self._token

    @token.setter
    def token(self, token: str):
        """
        Set the token for Authorization.

        Args:
            token (str): Token for authorization.
        """
        self._token = token

    def request_deserialize(
        self, method: str, path: str, res_model: Type[T], **kwargs
    ) -> T:
        """
        Make a request and deserialize the response to a defined type.

        Args:
            method (str): HTTP method (e.g. POST, GET, PUT, DELETE).
            path (str): URL path to call.
            res_model (Type[T]): Resulting type to which the response should be deserialized.
            **kwargs (optional): Optional keyword arguments to pass with the call to request.

        Returns:
            Parsed result of the call to the API, as res_model.
        """
        resp = self.request(method, path, **kwargs)
        return decode_response(res_model, resp)

    def request(
        self,
        method: str,
        path: str,
        endpoint: str = None,
        headers: Dict = None,
        **kwargs,
    ) -> Response:
        """
        Make a request.

        Args:
            method (str): HTTP method (e.g. POST, GET, PUT, DELETE).
            path (str): URL path to call.
            endpoint (str): Alternative endpoint for the AIS cluster (e.g. for connecting to a specific proxy).
            headers (dict): Extra headers to be passed with the request. Content-Type and User-Agent will be overridden.
            **kwargs (optional): Optional keyword arguments to pass with the call to request.

        Returns:
            Raw response from the API.
        """
        base = urljoin(endpoint, "v1") if endpoint else self._base_url
        url = f"{base}/{path.lstrip('/')}"
        if headers is None:
            headers = {}
        headers[HEADER_CONTENT_TYPE] = JSON_CONTENT_TYPE
        headers[HEADER_USER_AGENT] = f"{USER_AGENT_BASE}/{sdk_version}"
        if self.token:
            headers[HEADER_AUTHORIZATION] = f"Bearer {self.token}"

        resp = self.session_manager.session.request(
            method,
            url,
            headers=headers,
            timeout=self._timeout,
            **kwargs,
        )
        if resp.status_code < 200 or resp.status_code >= 300:
            handle_errors(resp, self._error_handler)
        return resp

    def get_full_url(self, path: str, params: Dict[str, Any]) -> str:
        """
        Get the full URL to the path on the cluster with the given parameters.

        Args:
            path: Path on the cluster.
            params: Query parameters to include.

        Returns:
            URL including cluster base URL and parameters.

        """
        return f"{self._base_url}/{path.lstrip('/')}?{urlencode(params)}"



# Returns the MD5 sum of the original data as the response.
# pylint: disable=unused-variable
MD5 = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-md5
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_md5:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 80
      command: ['/code/server.py', '--listen', '0.0.0.0', '--port', '80']
      readinessProbe:
        httpGet:
          path: /health
          port: default
"""

# Returns "Hello World!" on any request.
# pylint: disable=unused-variable
HELLO_WORLD = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-hello-world
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_hello_world:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 8000
      command: ["gunicorn", "main:app", "--workers", "20", "--worker-class", "uvicorn.workers.UvicornWorker", "--bind", "0.0.0.0:8000"]
      # command: ["uvicorn", "main:app", "--reload"]
      env:
        - name: ARG_TYPE
          value: "{arg_type}"
      readinessProbe:
        httpGet:
          path: /health
          port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# Returns "Hello World!" on any request.
# pylint: disable=unused-variable
GO_HELLO_WORLD = """
apiVersion: v1
kind: Pod
metadata:
  name: hello-world-go-transformer
  annotations:
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_hello_world_go:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 80
      command: ['./echo', '-l', '0.0.0.0', '-p', '80']
      readinessProbe:
        httpGet:
          path: /health
          port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# Returns the original data, with an MD5 sum in the response headers.
# pylint: disable=unused-variable
GO_ECHO = """
apiVersion: v1
kind: Pod
metadata:
  name: echo-go
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_echo_go:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 80
      command: ['./echo', '-l', '0.0.0.0', '-p', '80']
      readinessProbe:
        httpGet:
          path: /health
          port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# Returns the original data, with an MD5 sum in the response headers.
# pylint: disable=unused-variable
ECHO = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-echo
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_echo:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 8000
      command: ["gunicorn", "main:app", "--workers", "20", "--worker-class", "uvicorn.workers.UvicornWorker", "--bind", "0.0.0.0:8000"] 
      readinessProbe:
        httpGet:
          path: /health
          port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# Returns the transformed TensorFlow compatible data for the input TAR files. For
# more information on command options, visit
# https://github.com/NVIDIA/ais-etl/blob/master/transformers/tar2tf/README.md.
# pylint: disable=unused-variable
TAR2TF = """
apiVersion: v1
kind: Pod
metadata:
  name: tar2tf
  annotations:
    # Values it can take ["hpull://","hrev://","hpush://"]
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_tar2tf:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 80
      # To enable conversion e.g.
      command: ['./tar2tf', '-l', '0.0.0.0', '-p', '80', '{arg}', '{val}']
      readinessProbe:
        httpGet:
          path: /health
          port: default
"""

# Returns the compressed/decompressed file. For more information on command options, visit
# https://github.com/NVIDIA/ais-etl/blob/master/transformers/compress/README.md.
# pylint: disable=unused-variable
COMPRESS = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-compress
  annotations:
    # Values `communication_type` can take are ["hpull://", "hrev://", "hpush://", "io://"].
    # Visit https://github.com/NVIDIA/aistore/blob/main/docs/etl.md#communication-mechanisms
    # for more details.
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_compress:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 80
      command: ['/code/server.py', '--listen', '0.0.0.0', '--port', '80']
      env:
      # COMPRESS_OPTIONS is a dictionary (JSON string) of additional parameters
      # `mode` and `compression`. For more information on additional parameters, refer to
      # https://github.com/NVIDIA/ais-etl/blob/master/transformers/compress/README.md.
      - name: COMPRESS_OPTIONS
        value: '{compress_options}'
      readinessProbe:
        httpGet:
          path: /health
          port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# pylint: disable=unused-variable
KERAS_TRANSFORMER = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-keras
  annotations:
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_keras:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 8000
      command: ["gunicorn", "main:app", "--workers", "20", "--worker-class", "uvicorn.workers.UvicornWorker", "--bind", "0.0.0.0:8000"] 
      env:
        - name: FORMAT
          value: "{format}"
        - name: TRANSFORM
          value: '{transform}'
        - name: ARG_TYPE
          value: "{arg_type}"
      readinessProbe:
        httpGet:
          path: /health
          port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# Returns the FFMPEG decoded content. For more information on command options, visit
# https://github.com/NVIDIA/ais-etl/blob/master/transformers/ffmpeg/README.md.
# pylint: disable=unused-variable
FFMPEG = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-ffmpeg
  annotations:
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
  - name: server
    image: aistorage/transformer_ffmpeg:latest
    imagePullPolicy: Always
    ports:
    - name: default
      containerPort: 80
    command: ['/code/server.py', '--listen', '0.0.0.0', '--port', '80']
    env:
    # FFMPEG_OPTIONS is a dictionary (JSON string) of FFMPEG decode parameters. For more information on
    # FFMPEG decode parameters, refer to https://ffmpeg.org/ffmpeg.html#Synopsis.
    - name: FFMPEG_OPTIONS
      value: '{ffmpeg_options}'
    readinessProbe:
      httpGet:
        path: /health
        port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# Returns the transformed images using `Torchvision` pre-processing. For more
# information on command options, visit
# https://github.com/NVIDIA/ais-etl/blob/master/transformers/torchvision_preprocess/README.md.
# pylint: disable=unused-variable
TORCHVISION_TRANSFORMER = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-torchvision
  annotations:
    # Values `communication_type` can take are ["hpull://", "hrev://", "hpush://", "io://"].
    # Visit https://github.com/NVIDIA/aistore/blob/main/docs/etl.md#communication-mechanisms
    communication_type: "{communication_type}://"
    wait_timeout: 10m
spec:
  containers:
    - name: server
      image: aistorage/transformer_torchvision:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 8000
      command:  ["gunicorn", "main:app", "--workers", "20", "--worker-class", "uvicorn.workers.UvicornWorker", "--bind", "0.0.0.0:8000"]
      env:
        - name: FORMAT
        # expected values - PNG, JPEG, etc
          value: "{format}"
        - name: TRANSFORM
          value: '{transform}'
      readinessProbe:
        httpGet:
          path: /health
          port: default
      volumeMounts:
        - name: ais
          mountPath: /tmp/
  volumes:
    - name: ais
      hostPath:
        path: /tmp/
        type: Directory
"""

# pylint: disable=unused-variable
FACE_DETECTION_TRANSFORMER = """
apiVersion: v1
kind: Pod
metadata:
  name: transformer-face-detection
  annotations:
    communication_type: "{communication_type}://"
    wait_timeout: 5m
spec:
  containers:
    - name: server
      image: aistorage/transformer_face_detection:latest
      imagePullPolicy: Always
      ports:
        - name: default
          containerPort: 8000
      command:  ["gunicorn", "main:app", "--workers", "5", "--worker-class", "uvicorn.workers.UvicornWorker", "--bind", "0.0.0.0:8000"]
      readinessProbe:
        httpGet:
          path: /health
          port: default
      env:
        - name: FORMAT
          value: "{format}"
        - name: ARG_TYPE
          value: "{arg_type}"
      volumeMounts:
        - name: ais
          mountPath: /tmp/ais
  volumes:
    - name: ais
      hostPath:
        path: /tmp/ais
        type: Directory
"""



#
# Copyright (c) 2021-2024, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
import base64
from dataclasses import dataclass
from typing import Any, Mapping, List, Optional, Dict

import msgspec
import humanize
from pydantic import BaseModel, Field, validator

from requests.structures import CaseInsensitiveDict

from aistore.sdk.namespace import Namespace
from aistore.sdk.archive_mode import ArchiveMode
from aistore.sdk.list_object_flag import ListObjectFlag
from aistore.sdk.object_props import ObjectProps
from aistore.sdk.const import (
    HEADER_CONTENT_LENGTH,
    AIS_CHECKSUM_VALUE,
    AIS_ACCESS_TIME,
    AIS_VERSION,
    AIS_OBJ_NAME,
    AIS_LOCATION,
    AIS_MIRROR_COPIES,
    PROVIDER_AIS,
)


# pylint: disable=too-few-public-methods,unused-variable,missing-function-docstring


class ActionMsg(BaseModel):
    """
    Represents the action message passed by the client via json
    """

    action: str
    name: str = ""
    value: Any = None


class NetInfo(BaseModel):
    """
    Represents a set of network-related info
    """

    node_hostname: str = ""
    daemon_port: str = ""
    direct_url: str = ""


class Snode(BaseModel):
    """
    Represents a system node
    """

    daemon_id: str
    daemon_type: str
    public_net: NetInfo = None
    intra_control_net: NetInfo = None
    intra_data_net: NetInfo = None
    flags: int = 0


class Smap(BaseModel):
    """
    Represents a system map
    """

    tmap: Mapping[str, Snode]
    pmap: Mapping[str, Snode]
    proxy_si: Snode
    version: int = 0
    uuid: str = ""
    creation_time: str = ""


class BucketEntry(msgspec.Struct):
    """
    Represents a single entry in a bucket -- an object
    See cmn/objlist.go/LsoEntry
    """

    n: str
    cs: str = ""
    a: str = ""
    v: str = ""
    t: str = ""
    s: int = 0
    c: int = 0
    f: int = 0
    object: Any = None

    def generate_object_props(self) -> ObjectProps:
        """
        Convert bucket entry data into Object Props.

        NOTE: Bucket entry data and object props are not a one-to-one mapping.

        Returns:
            ObjectProps with object data
        """
        headers = CaseInsensitiveDict(
            {
                AIS_OBJ_NAME: self.name,
                AIS_CHECKSUM_VALUE: self.checksum,
                AIS_ACCESS_TIME: self.atime,
                AIS_VERSION: self.version,
                AIS_LOCATION: self.location,
                HEADER_CONTENT_LENGTH: self.size,
                AIS_MIRROR_COPIES: self.copies,
            }
        )
        return ObjectProps(headers)

    @property
    def name(self):
        return self.n

    @property
    def checksum(self):
        return self.cs

    @property
    def atime(self):
        return self.a

    @property
    def version(self):
        return self.v

    @property
    def location(self):
        return self.t

    @property
    def size(self):
        return self.s

    @property
    def copies(self):
        return self.c

    @property
    def flags(self):
        return self.f

    def is_cached(self):
        return (self.flags & (1 << 6)) != 0

    def is_ok(self):
        return (self.flags & ((1 << 5) - 1)) == 0


class BucketList(msgspec.Struct):
    """
    Represents the response when getting a list of bucket items, containing a list of BucketEntry objects
    """

    UUID: str
    ContinuationToken: str
    Flags: int
    Entries: List[BucketEntry] = None

    @property
    def uuid(self):
        return self.UUID

    @property
    def continuation_token(self):
        return self.ContinuationToken

    @property
    def flags(self):
        return self.Flags

    @property
    def entries(self):
        return [] if self.Entries is None else self.Entries

    def get_entries(self):
        """
        Deprecated -- use entries property
        """
        return self.entries


class BucketModel(BaseModel):
    """
    Represents the response from the API containing bucket info
    """

    name: str
    provider: str = PROVIDER_AIS
    namespace: Namespace = None

    def as_dict(self):
        dict_rep = {"name": self.name, "provider": self.provider}
        if self.namespace:
            dict_rep["namespace"] = self.namespace
        return dict_rep


class BsummCtrlMsg(BaseModel):
    """
    Represents the bucket summary control message
    """

    uuid: str
    prefix: str
    cached: bool
    present: bool


class JobArgs(BaseModel):
    """
    Represents the set of args to pass when making a job-related request
    """

    id: str = ""
    kind: str = ""
    daemon_id: str = ""
    bucket: BucketModel = None
    buckets: List[BucketModel] = None

    def as_dict(self):
        return {
            "ID": self.id,
            "Kind": self.kind,
            "DaemonID": self.daemon_id,
            "Bck": self.bucket,
            "Buckets": self.buckets,
        }


class JobQuery(BaseModel):
    """
    Structure to send the API when querying the cluster for multiple jobs
    """

    active: bool = False
    kind: str = ""
    target: str = ""

    def as_dict(self):
        return {
            "kind": self.kind,
            "node": self.target,
            "show_active": self.active,
        }


class JobStatus(BaseModel):
    """
    Represents the response of an API query to fetch job status
    """

    uuid: str = ""
    err: str = ""
    end_time: int = 0
    aborted: bool = False


class ETLInfo(BaseModel):  # pylint: disable=too-few-public-methods,unused-variable
    """
    Represents the API response when querying an ETL
    """

    id: str = ""
    xaction_id: str = ""
    obj_count: int = 0
    in_bytes: int = 0
    out_bytes: int = 0


class ETLDetails(BaseModel):
    """
    Represents the API response of queries on single ETL details
    """

    id: str
    communication: str
    timeout: str
    code: Optional[bytes]
    spec: Optional[str]
    dependencies: Optional[str]
    runtime: Optional[str]  # see ext/etl/runtime/all.go
    chunk_size: int = 0
    argument: str = ""

    @validator("code")
    def set_code(cls, code):  # pylint: disable=no-self-argument
        if code is not None:
            code = base64.b64decode(code)
        return code

    @validator("spec")
    def set_spec(cls, spec):  # pylint: disable=no-self-argument
        if spec is not None:
            spec = base64.b64decode(spec)
        return spec


class InitETLArgs(BaseModel):
    """
    Represents the args shared by ETL initialization with code or spec
    """

    etl_name: str
    communication_type: str
    timeout: str
    arg_type: str = ""


class InitSpecETLArgs(InitETLArgs):
    """
    Represents the set of args the sdk will pass to AIStore when making a request to initialize an ETL with a spec
    """

    spec: str

    def as_dict(self):
        return {
            "id": self.etl_name,
            "timeout": self.timeout,
            "communication": f"{self.communication_type}://",
            "spec": self.spec,
            "argument": self.arg_type,
        }


class InitCodeETLArgs(InitETLArgs):
    """
    Represents the set of args the sdk will pass to AIStore when making a request to initialize an ETL with code
    """

    runtime: str
    dependencies: str
    functions: Dict[str, str]
    code: str
    chunk_size: int = None

    def as_dict(self):
        dict_rep = {
            "id": self.etl_name,
            "runtime": self.runtime,
            "communication": f"{self.communication_type}://",
            "timeout": self.timeout,
            "funcs": self.functions,
            "code": self.code,
            "dependencies": self.dependencies,
            "argument": self.arg_type,
        }
        if self.chunk_size:
            dict_rep["chunk_size"] = self.chunk_size
        return dict_rep


class PromoteAPIArgs(BaseModel):
    """
    Represents the set of args the sdk will pass to AIStore when making a promote request and
    provides conversion to the expected json format
    """

    target_id: str = ""
    source_path: str = ""
    object_name: str = ""
    recursive: bool = False
    overwrite_dest: bool = False
    delete_source: bool = False
    src_not_file_share: bool = False

    def as_dict(self):
        return {
            "tid": self.target_id,
            "src": self.source_path,
            "obj": self.object_name,
            "rcr": self.recursive,
            "ovw": self.overwrite_dest,
            "dls": self.delete_source,
            "notshr": self.src_not_file_share,
        }


class JobStats(BaseModel):
    """
    Structure for job statistics
    """

    objects: int = 0
    bytes: int = 0
    out_objects: int = 0
    out_bytes: int = 0
    in_objects: int = 0
    in_bytes: int = 0


class JobSnapshot(BaseModel):
    """
    Structure for the data returned when querying a single job on a single target node
    """

    id: str = Field(default="", alias="id")
    kind: str = Field(default="", alias="kind")
    start_time: str = Field(default="", alias="start-time")
    end_time: str = Field(default="", alias="end-time")
    bucket: BucketModel = Field(default=None, alias="bck")
    source_bck: str = Field(default="", alias="source-bck")
    dest_bck: str = Field(default="", alias="dest-bck")
    rebalance_id: str = Field(default="", alias="glob.id")
    stats: JobStats = Field(default=None, alias="stats")
    aborted: bool = Field(default=False, alias="aborted")
    is_idle: bool = Field(default=False, alias="is_idle")
    abort_err: str = Field(default="", alias="abort-err")

    class Config:
        """
        Configuration for Pydantic model to use alias.
        """

        allow_population_by_field_name = True


class CopyBckMsg(BaseModel):
    """
    API message structure for copying a bucket
    """

    prefix: str = ""
    prepend: str
    dry_run: bool
    force: bool
    latest: bool
    sync: bool

    def as_dict(self):
        return {
            "prefix": self.prefix,
            "prepend": self.prepend,
            "dry_run": self.dry_run,
            "force": self.force,
            "latest-ver": self.latest,
            "synchronize": self.sync,
        }


class ListObjectsMsg(BaseModel):
    """
    API message structure for listing objects in a bucket
    """

    prefix: str
    page_size: int
    uuid: str
    props: str
    continuation_token: str
    flags: List[ListObjectFlag]
    target: str

    def as_dict(self):
        return {
            "prefix": self.prefix,
            "pagesize": self.page_size,
            "uuid": self.uuid,
            "props": self.props,
            "continuation_token": self.continuation_token,
            "flags": str(ListObjectFlag.join_flags(self.flags)),
            "target": self.target,
        }


class TransformBckMsg(BaseModel):
    """
    API message structure for requesting an etl transform on a bucket
    """

    etl_name: str
    timeout: str

    def as_dict(self):
        return {"id": self.etl_name, "request_timeout": self.timeout}


class TCBckMsg(BaseModel):
    """
    API message structure for transforming or copying between buckets.
    Can be used on its own for an entire bucket or encapsulated in TCMultiObj to apply only to a selection of objects
    """

    ext: Dict[str, str] = None
    copy_msg: CopyBckMsg = None
    transform_msg: TransformBckMsg = None

    def as_dict(self):
        dict_rep = {}
        if self.ext:
            dict_rep["ext"] = self.ext
        if self.copy_msg:
            for key, val in self.copy_msg.as_dict().items():
                dict_rep[key] = val
        if self.transform_msg:
            for key, val in self.transform_msg.as_dict().items():
                dict_rep[key] = val
        return dict_rep


class PrefetchMsg(BaseModel):
    """
    API message structure for prefetching objects from remote buckets.
    """

    object_selection: Dict
    continue_on_err: bool
    latest: bool
    blob_threshold: int = None

    def as_dict(self):
        dict_rep = self.object_selection
        dict_rep["coer"] = self.continue_on_err
        dict_rep["latest-ver"] = self.latest
        if self.blob_threshold:
            dict_rep["blob-threshold"] = self.blob_threshold
        return dict_rep


class TCMultiObj(BaseModel):
    """
    API message structure for transforming or copying multiple objects between buckets
    """

    to_bck: BucketModel
    tc_msg: TCBckMsg = None
    continue_on_err: bool
    object_selection: Dict

    def as_dict(self):
        dict_rep = self.object_selection
        if self.tc_msg:
            for key, val in self.tc_msg.as_dict().items():
                dict_rep[key] = val
        dict_rep["tobck"] = self.to_bck.as_dict()
        dict_rep["coer"] = self.continue_on_err
        return dict_rep


class ArchiveMultiObj(BaseModel):
    """
    API message structure for multi-object archive requests
    """

    archive_name: str
    to_bck: BucketModel
    mime: str = None
    include_source_name = False
    allow_append = False
    continue_on_err = False
    object_selection: Dict

    def as_dict(self):
        dict_rep = self.object_selection
        dict_rep["archname"] = self.archive_name
        dict_rep["isbn"] = self.include_source_name
        dict_rep["aate"] = self.allow_append
        dict_rep["coer"] = self.continue_on_err
        if self.mime:
            dict_rep["mime"] = self.mime
        if self.to_bck:
            dict_rep["tobck"] = self.to_bck.as_dict()
        return dict_rep


class BlobMsg(BaseModel):
    """
    Represents the set of args the sdk will pass to AIStore when making a blob-download request
    and provides conversion to the expected json format
    """

    chunk_size: int = None
    num_workers: int = None
    latest: bool

    def as_dict(self):
        dict_rep = {
            "latest-ver": self.latest,
        }
        if self.chunk_size:
            dict_rep["chunk-size"] = self.chunk_size
        if self.num_workers:
            dict_rep["num-workers"] = self.num_workers
        return dict_rep


class NodeTracker(BaseModel):
    """
    Represents the tracker info of a cluster node
    """

    append_ns: int = Field(0, alias="append.ns")
    del_n: int = Field(0, alias="del.n")
    disk_sdb_util: float = Field(0, alias="disk.sdb..util")
    disk_sdb_avg_rsize: float = Field(0, alias="disk.sdb.avg.rsize")
    disk_sdb_avg_wsize: float = Field(0, alias="disk.sdb.avg.wsize")
    disk_sdb_read_bps: int = Field(0, alias="disk.sdb.read.bps")
    disk_sdb_write_bps: int = Field(0, alias="disk.sdb.write.bps")
    dl_ns: int = Field(0, alias="dl.ns")
    dsort_creation_req_n: int = Field(0, alias="dsort.creation.req.n")
    dsort_creation_resp_n: int = Field(0, alias="dsort.creation.resp.n")
    dsort_creation_resp_ns: int = Field(0, alias="dsort.creation.resp.ns")
    dsort_extract_shard_mem_n: int = Field(0, alias="dsort.extract.shard.mem.n")
    dsort_extract_shard_size: int = Field(0, alias="dsort.extract.shard.size")
    err_del_n: int = Field(0, alias="err.del.n")
    err_get_n: int = Field(0, alias="err.get.n")
    get_bps: int = Field(0, alias="get.bps")
    get_cold_n: int = Field(0, alias="get.cold.n")
    get_cold_rw_ns: int = Field(0, alias="get.cold.rw.ns")
    get_cold_size: int = Field(0, alias="get.cold.size")
    get_n: int = Field(0, alias="get.n")
    get_ns: int = Field(0, alias="get.ns")
    get_redir_ns: int = Field(0, alias="get.redir.ns")
    get_size: int = Field(0, alias="get.size")
    kalive_ns: int = Field(0, alias="kalive.ns")
    lcache_evicted_n: int = Field(0, alias="lcache.evicted.n")
    lcache_flush_cold_n: int = Field(0, alias="lcache.flush.cold.n")
    lru_evict_n: int = Field(0, alias="lru.evict.n")
    lru_evict_size: int = Field(0, alias="lru.evict.size")
    lst_n: int = Field(0, alias="lst.n")
    lst_ns: int = Field(0, alias="lst.ns")
    put_bps: int = Field(0, alias="put.bps")
    put_n: int = Field(0, alias="put.n")
    put_ns: int = Field(0, alias="put.ns")
    put_redir_ns: int = Field(0, alias="put.redir.ns")
    put_size: int = Field(0, alias="put.size")
    remote_deleted_del_n: int = Field(0, alias="remote.deleted.del.n")
    stream_in_n: int = Field(0, alias="stream.in.n")
    stream_in_size: int = Field(0, alias="stream.in.size")
    stream_out_n: int = Field(0, alias="stream.out.n")
    stream_out_size: int = Field(0, alias="stream.out.size")
    up_ns_time: int = Field(0, alias="up.ns.time")
    ver_change_n: int = Field(0, alias="ver.change.n")
    ver_change_size: int = Field(0, alias="ver.change.size")

    class Config:
        """
        Configuration for Pydantic model to use alias.
        """

        allow_population_by_field_name = True


class DiskInfo(BaseModel):
    """
    Represents the disk info of a node mountpath
    """

    used: str
    avail: str
    pct_used: float
    disks: List[str]
    mountpath_label: str
    fs: str


class DiskInfoV322(BaseModel):
    """
    Represents the disk info of a node mountpath
    """

    used: str
    avail: str
    pct_used: float
    disks: List[str]
    fs: str


class NodeCapacityV322(BaseModel):
    """
    Represents the capacity info of a node, including its mountpaths.
    """

    mountpaths: Mapping[str, DiskInfoV322] = Field({}, alias="MountPaths")
    pct_max: float
    pct_avg: float
    pct_min: float
    cs_err: str

    class Config:
        """
        Configuration for Pydantic model to use alias.
        """

        allow_population_by_field_name = True


class NodeCapacity(BaseModel):
    """
    Represents the capacity info of a node, including its mountpaths.
    """

    mountpaths: Mapping[str, DiskInfo] = Field({}, alias="MountPaths")
    pct_max: float
    pct_avg: float
    pct_min: float
    cs_err: str
    total_used: int
    total_avail: int

    class Config:
        """
        Configuration for Pydantic model to use alias.
        """

        allow_population_by_field_name = True


class NodeStatsV322(BaseModel):
    """
    Represents the response from cluster performance API
    """

    snode: Snode
    tracker: NodeTracker
    capacity: NodeCapacityV322
    rebalance_snap: Optional[Dict] = None
    status: str
    deployment: str
    ais_version: str
    build_time: str
    k8s_pod_name: str
    sys_info: Dict
    smap_version: str


class NodeStats(BaseModel):
    """
    Represents the response from cluster performance API
    """

    snode: Snode
    tracker: NodeTracker
    capacity: NodeCapacity
    rebalance_snap: Optional[Dict] = None
    status: str
    deployment: str
    ais_version: str
    build_time: str
    k8s_pod_name: str
    sys_info: Dict
    smap_version: str
    reserved1: Optional[str] = ""
    reserved2: Optional[str] = ""
    reserved3: Optional[int] = 0
    reserved4: Optional[int] = 0


# pylint: disable=too-many-instance-attributes
class NodeThroughput:
    """
    Represents the throughput stats of a node
    """

    def __init__(self, node_tracker: NodeTracker):
        get_cold_avg_size = 0
        if node_tracker.get_cold_n > 0:
            get_cold_avg_size = node_tracker.get_cold_size / node_tracker.get_cold_n

        get_avg_size = 0
        if node_tracker.get_n > 0:
            get_avg_size = node_tracker.get_bps / node_tracker.get_n

        put_avg_size = 0
        if node_tracker.put_n > 0:
            put_avg_size = node_tracker.put_bps / node_tracker.put_n

        self.get_bw = node_tracker.get_bps
        self.get_cold_n = node_tracker.get_cold_n
        self.get_cold_total_size = node_tracker.get_cold_size
        self.get_cold_avg_size = get_cold_avg_size
        self.get_n = node_tracker.get_n
        self.get_total_size = node_tracker.get_bps
        self.get_avg_size = get_avg_size
        self.put_bw = node_tracker.put_bps
        self.put_n = node_tracker.put_n
        self.put_total_size = node_tracker.put_bps
        self.put_avg_size = put_avg_size
        self.err_get_n = node_tracker.err_get_n

    def as_dict(self):
        return {
            "GET(bw)": f"{humanize.naturalsize(self.get_bw)}/s",
            "GET-COLD(n)": self.get_cold_n,
            "GET-COLD(total size)": humanize.naturalsize(self.get_cold_total_size),
            "GET-COLD(avg size)": humanize.naturalsize(self.get_cold_avg_size),
            "GET(n)": self.get_n,
            "GET(total size)": humanize.naturalsize(self.get_total_size),
            "GET(avg size)": humanize.naturalsize(self.get_avg_size),
            "PUT(bw)": f"{humanize.naturalsize(self.put_bw)}/s",
            "PUT(n)": self.put_n,
            "PUT(total size)": humanize.naturalsize(self.put_total_size),
            "PUT(avg size)": humanize.naturalsize(self.put_avg_size),
            "ERR-GET(n)": self.err_get_n,
        }


# pylint: disable=too-many-instance-attributes
class NodeLatency:
    """
    Represents the latency stats of a node
    """

    def __init__(self, node_tracker: NodeTracker):
        get_cold_avg_size = 0
        if node_tracker.get_cold_n > 0:
            get_cold_avg_size = node_tracker.get_cold_size / node_tracker.get_cold_n

        get_avg_size = 0
        if node_tracker.get_n > 0:
            get_avg_size = node_tracker.get_bps / node_tracker.get_n

        put_avg_size = 0
        if node_tracker.put_n > 0:
            put_avg_size = node_tracker.put_bps / node_tracker.put_n

        self.get_cold_n = node_tracker.get_cold_n
        self.get_cold_total_size = node_tracker.get_cold_size
        self.get_cold_avg_size = get_cold_avg_size
        self.get_n = node_tracker.get_n
        self.get_total_size = node_tracker.get_bps
        self.get_avg_size = get_avg_size
        self.put_n = node_tracker.put_n
        self.put_total_size = node_tracker.put_bps
        self.put_avg_size = put_avg_size
        self.err_get_n = node_tracker.err_get_n

    def as_dict(self):
        return {
            "GET-COLD(n)": self.get_cold_n,
            "GET-COLD(total size)": humanize.naturalsize(self.get_cold_total_size),
            "GET-COLD(avg size)": humanize.naturalsize(self.get_cold_avg_size),
            "GET(n)": self.get_n,
            "GET(total size)": humanize.naturalsize(self.get_total_size),
            "GET(avg size)": humanize.naturalsize(self.get_avg_size),
            "PUT(n)": self.put_n,
            "PUT(total size)": humanize.naturalsize(self.put_total_size),
            "PUT(avg size)": humanize.naturalsize(self.put_avg_size),
            "ERR-GET(n)": self.err_get_n,
        }


# pylint: disable=too-many-instance-attributes
class NodeCounter:
    """
    Represents the counter stats of a node
    """

    def __init__(self, node_tracker: NodeTracker):
        self.del_n = node_tracker.del_n
        self.dsort_creation_req_n = node_tracker.dsort_creation_req_n
        self.dsort_creation_resp_n = node_tracker.dsort_creation_resp_n
        self.dsort_extract_shard_mem_n = node_tracker.dsort_extract_shard_mem_n
        self.dsort_extract_shard_size = node_tracker.dsort_extract_shard_size
        self.get_cold_n = node_tracker.get_cold_n
        self.get_cold_size = node_tracker.get_cold_size
        self.get_n = node_tracker.get_n
        self.get_size = node_tracker.get_size
        self.lcache_evicted_n = node_tracker.lcache_evicted_n
        self.lcache_flush_cold_n = node_tracker.lcache_flush_cold_n
        self.evict_n = node_tracker.lru_evict_n
        self.evict_size = node_tracker.lru_evict_size
        self.list_n = node_tracker.lst_n
        self.put_n = node_tracker.put_n
        self.put_size = node_tracker.put_size
        self.remote_deleted_del_n = node_tracker.remote_deleted_del_n
        self.stream_in_n = node_tracker.stream_in_n
        self.stream_in_size = node_tracker.stream_in_size
        self.stream_out_n = node_tracker.stream_out_n
        self.stream_out_size = node_tracker.stream_out_size
        self.version_change_n = node_tracker.ver_change_n
        self.version_change_size = node_tracker.ver_change_size
        self.err_del_n = node_tracker.err_del_n
        self.err_get_n = node_tracker.err_get_n

    def as_dict(self):
        return {
            "DELETE(n)": self.del_n,
            "DSORT-CREATION-REQ(n)": self.dsort_creation_req_n,
            "DSORT-CREATION-RESP(n)": self.dsort_creation_resp_n,
            "DSORT-EXTRACT-SHARD-MEM(n)": self.dsort_extract_shard_mem_n,
            "DSORT-EXTRACT-SHARD(size)": humanize.naturalsize(
                self.dsort_extract_shard_size
            ),
            "GET-COLD(n)": self.get_cold_n,
            "GET-COLD(size)": humanize.naturalsize(self.get_cold_size),
            "GET(n)": self.get_n,
            "GET(size)": humanize.naturalsize(self.get_size),
            "LCACHE-EVICTED(n)": self.lcache_evicted_n,
            "LCACHE-FLUSH-COLD(n)": self.lcache_flush_cold_n,
            "EVICT(n)": self.evict_n,
            "EVICT(size)": humanize.naturalsize(self.evict_size),
            "LIST(n)": self.list_n,
            "PUT(n)": self.put_n,
            "PUT(size)": humanize.naturalsize(self.put_size),
            "REMOTE-DELETED-DEL(n)": self.remote_deleted_del_n,
            "STREAM-IN(n)": self.stream_in_n,
            "STREAM-IN(size)": humanize.naturalsize(self.stream_in_size),
            "STREAM-OUT(n)": self.stream_out_n,
            "STREAM-OUT(size)": humanize.naturalsize(self.stream_out_size),
            "VERSION-CHANGE(n)": self.version_change_n,
            "VERSION-CHANGE(size)": humanize.naturalsize(self.version_change_size),
            "ERR-DEL(n)": self.err_del_n,
            "ERR-GET(n)": self.err_get_n,
        }


class ClusterPerformance:
    """
    Represents the performance metrics for the cluster
    """

    def __init__(
        self,
        throughput: Mapping[str, NodeThroughput],
        latency: Mapping[str, NodeLatency],
        counters: Mapping[str, NodeCounter],
    ):
        self.throughput = throughput
        self.latency = latency
        self.counters = counters

    def as_dict(self):
        return {
            "throughput": self.throughput,
            "latency": self.latency,
            "counters": self.counters,
        }


@dataclass
class ArchiveSettings:
    """
    Configuration for extracting files from an archive

    Attributes:
    archpath (str, optional): If the object is an archive, use `archpath` to extract a single file
        from the archive
    regex (str, optional): A prefix, suffix, WebDataset key, or general-purpose regular expression
        used to match filenames within the archive and select possibly multiple files
    mode (ArchiveMode, optional): Specifies the mode of archive extraction when using `regex`

    Example:
        # Extract a single file from an archive
        single_file_settings = ArchiveSettings(
            archpath="path/to/your/file.txt"
        )

        # Extract multiple files from an archive
        multi_file_settings = ArchiveSettings(
            regex = "log",  # Retrieve all log files from the archive
            mode=ArchiveMode.SUFFIX,
        )
    """

    archpath: str = ""
    regex: str = ""
    mode: ArchiveMode = None

    def __post_init__(self):
        if self.mode and not self.regex:
            raise ValueError("Archive mode requires archive regex")

        if self.regex and not self.mode:
            raise ValueError("Archive regex requires archive mode")

        if self.regex and self.archpath:
            raise ValueError("Cannot use both Archive regex and Archive path")


@dataclass
class BlobDownloadSettings:
    """
    Configuration for downloading objects using a blob downloader

    Attributes:
        chunk_size (str, optional): Chunk size for the blob downloader. It can be specified in IEC
            or SI units, or as raw bytes (e.g., "4mb", "1MiB", "1048576", "128k")
        num_workers (str, optional): Number of concurrent workers for the blob downloader

    Example:
        blob_settings = BlobDownloadSettings(
            chunk_size="1MiB",  # 1 MiB per chunk
            num_workers="5"     # 5 concurrent download workers
        )
    """

    chunk_size: str = None
    num_workers: str = None



#
# Copyright (c) 2018-2024, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
from typing import Optional, Tuple, Union

from urllib3 import Retry

from aistore.sdk.bucket import Bucket
from aistore.sdk.const import (
    PROVIDER_AIS,
)
from aistore.sdk.cluster import Cluster
from aistore.sdk.dsort import Dsort
from aistore.sdk.request_client import RequestClient
from aistore.sdk.session_manager import SessionManager
from aistore.sdk.types import Namespace
from aistore.sdk.job import Job
from aistore.sdk.etl import Etl
from aistore.sdk.utils import parse_url
from aistore.sdk.object import Object
from aistore.sdk.errors import InvalidURLException


# pylint: disable=unused-variable, duplicate-code, too-many-arguments
class Client:
    """
    AIStore client for managing buckets, objects, ETL jobs

    Args:
        endpoint (str): AIStore endpoint
        skip_verify (bool, optional): If True, skip SSL certificate verification. Defaults to False.
        ca_cert (str, optional): Path to a CA certificate file for SSL verification.
        timeout (Union[float, Tuple[float, float], None], optional): Request timeout in seconds; a single float
            for both connect/read timeouts (e.g., 5.0), a tuple for separate connect/read timeouts (e.g., (3.0, 10.0)),
            or None to disable timeout.
        retry (urllib3.Retry, optional): Retry configuration object from the urllib3 library.
        token (str, optional): Authorization token.
    """

    def __init__(
        self,
        endpoint: str,
        skip_verify: bool = False,
        ca_cert: str = None,
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        retry: Retry = None,
        token: str = None,
    ):
        session_manager = SessionManager(
            retry=retry, ca_cert=ca_cert, skip_verify=skip_verify
        )
        self._request_client = RequestClient(
            endpoint=endpoint,
            session_manager=session_manager,
            timeout=timeout,
            token=token,
        )

    def bucket(
        self, bck_name: str, provider: str = PROVIDER_AIS, namespace: Namespace = None
    ):
        """
        Factory constructor for bucket object.
        Does not make any HTTP request, only instantiates a bucket object.

        Args:
            bck_name (str): Name of bucket
            provider (str): Provider of bucket, one of "ais", "aws", "gcp", ... (optional, defaults to ais)
            namespace (Namespace): Namespace of bucket (optional, defaults to None)

        Returns:
            The bucket object created.
        """
        return Bucket(
            client=self._request_client,
            name=bck_name,
            provider=provider,
            namespace=namespace,
        )

    def cluster(self):
        """
        Factory constructor for cluster object.
        Does not make any HTTP request, only instantiates a cluster object.

        Returns:
            The cluster object created.
        """
        return Cluster(client=self._request_client)

    def job(self, job_id: str = "", job_kind: str = ""):
        """
        Factory constructor for job object, which contains job-related functions.
        Does not make any HTTP request, only instantiates a job object.

        Args:
            job_id (str, optional): Optional ID for interacting with a specific job
            job_kind (str, optional): Optional specific type of job empty for all kinds

        Returns:
            The job object created.
        """
        return Job(client=self._request_client, job_id=job_id, job_kind=job_kind)

    def etl(self, etl_name: str):
        """
        Factory constructor for ETL object.
        Contains APIs related to AIStore ETL operations.
        Does not make any HTTP request, only instantiates an ETL object.

        Args:
            etl_name (str): Name of the ETL

        Returns:
            The ETL object created.
        """
        return Etl(client=self._request_client, name=etl_name)

    def dsort(self, dsort_id: str = ""):
        """
        Factory constructor for dSort object.
        Contains APIs related to AIStore dSort operations.
        Does not make any HTTP request, only instantiates a dSort object.

        Args:
            dsort_id: ID of the dSort job

        Returns:
            dSort object created
        """
        return Dsort(client=self._request_client, dsort_id=dsort_id)

    def fetch_object_by_url(self, url: str) -> Object:
        """
        Retrieve an object based on its URL.

        Args:
            url (str): Full URL of the object (e.g., "ais://bucket1/file.txt")

        Returns:
            Object: The object retrieved from the specified URL
        """
        try:
            provider, bck_name, obj_name = parse_url(url)
            if not provider or not bck_name or not obj_name:
                raise InvalidURLException(url)
            return self.bucket(bck_name, provider=provider).object(obj_name)
        except InvalidURLException as err:
            raise err



"""
Multishard Stream Dataset for AIS.

Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
"""

from aistore.sdk.dataset.data_shard import DataShard
from aistore.sdk import Bucket
from typing import Iterator, List, Iterable
from aistore.sdk.list_object_flag import ListObjectFlag
from aistore.sdk.types import ArchiveSettings
from torch.utils.data import IterableDataset


class AISMultiShardStream(IterableDataset):
    """
    An iterable-style dataset that iterates over multiple shard streams and yields combined samples.

    Args:
        data_sources (List[DataShard]): List of DataShard objects

    Returns:
        Iterable: Iterable over the combined samples, where each sample is a tuple of
            one object bytes from each shard stream
    """

    def __init__(self, data_sources: List[DataShard]):
        self._data_sources = data_sources

    def __iter__(self) -> Iterator:
        data_iterators = (
            self._get_shard_objects_iterator(ds.bucket, ds.prefix, ds.etl_name)
            for ds in self._data_sources
        )
        return zip(*data_iterators)

    def _get_shard_objects_iterator(
        self, bucket: Bucket, prefix: str = "", etl_name: str = ""
    ) -> Iterable[bytes]:
        """
        Create an iterable over all the objects in the given shards.

        Args:
            bucket (Bucket): Bucket containing the shards
            prefix (str): Prefix of the object names
            etl_name (str): ETL name to apply on each object

        Returns:
            Iterable[Object]: Iterable over all the objects in the given shards,
                            with each iteration returning a combined sample
        """
        shards_iter = bucket.list_objects_iter(prefix=prefix, props="name")

        for shard in shards_iter:
            path = shard.name
            objects_iter = bucket.list_objects_iter(
                prefix=path, props="name", flags=[ListObjectFlag.ARCH_DIR]
            )

            for obj in objects_iter:
                if obj.name != path:
                    obj_name = obj.name.replace(f"{path}/", "", 1)
                    yield bucket.object(path).get(
                        etl_name=etl_name,
                        archive_settings=ArchiveSettings(archpath=obj_name),
                    ).read_all()



#
# Copyright (c) 2022-2024, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable

import logging
from typing import List, Optional

from aistore.sdk.const import (
    HTTP_METHOD_GET,
    ACT_LIST,
    PROVIDER_AIS,
    QPARAM_WHAT,
    QPARAM_PRIMARY_READY_REB,
    QPARAM_PROVIDER,
    HEADER_NODE_ID,
    URL_PATH_ETL,
    URL_PATH_REVERSE,
    URL_PATH_BUCKETS,
    URL_PATH_HEALTH,
    URL_PATH_DAEMON,
    URL_PATH_CLUSTER,
    WHAT_SMAP,
    WHAT_ALL_XACT_STATUS,
    WHAT_ALL_RUNNING_STATUS,
    WHAT_NODE_STATS_AND_STATUS,
    WHAT_NODE_STATS_AND_STATUS_V322,
)

from aistore.sdk.types import (
    BucketModel,
    JobStatus,
    JobQuery,
    ETLInfo,
    ActionMsg,
    Smap,
    NodeStats,
    NodeStatsV322,
    ClusterPerformance,
    NodeCounter,
    NodeLatency,
    NodeThroughput,
)
from aistore.sdk.request_client import RequestClient
from aistore.sdk.errors import AISError

logger = logging.getLogger("cluster")


# pylint: disable=unused-variable
class Cluster:
    """
    A class representing a cluster bound to an AIS client.
    """

    # pylint: disable=duplicate-code
    def __init__(self, client: RequestClient):
        self._client = client

    @property
    def client(self):
        """Client this cluster uses to make requests"""
        return self._client

    def get_info(self) -> Smap:
        """
        Returns state of AIS cluster, including the detailed information about its nodes.

        Returns:
            aistore.sdk.types.Smap: Smap containing cluster information

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        return self._get_smap()

    def get_primary_url(self) -> str:
        """
        Returns: URL of primary proxy
        """
        return self._get_smap().proxy_si.public_net.direct_url

    def list_buckets(self, provider: str = PROVIDER_AIS):
        """
        Returns list of buckets in AIStore cluster.

        Args:
            provider (str, optional): Name of bucket provider, one of "ais", "aws", "gcp", "az" or "ht".
            Defaults to "ais". Empty provider returns buckets of all providers.

        Returns:
            List[BucketModel]: A list of buckets

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        params = {QPARAM_PROVIDER: provider}
        action = ActionMsg(action=ACT_LIST).dict()

        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_BUCKETS,
            res_model=List[BucketModel],
            json=action,
            params=params,
        )

    def list_jobs_status(self, job_kind="", target_id="") -> List[JobStatus]:
        """
        List the status of jobs on the cluster

        Args:
            job_kind (str, optional): Only show jobs of a particular type
            target_id (str, optional): Limit to jobs on a specific target node

        Returns:
            List of JobStatus objects
        """
        res = self._client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_CLUSTER,
            res_model=Optional[List[JobStatus]],
            json=JobQuery(kind=job_kind, target=target_id).as_dict(),
            params={QPARAM_WHAT: WHAT_ALL_XACT_STATUS},
        )
        if res is None:
            return []
        return res

    def list_running_jobs(self, job_kind="", target_id="") -> List[str]:
        """
        List the currently running jobs on the cluster

        Args:
            job_kind (str, optional): Only show jobs of a particular type
            target_id (str, optional): Limit to jobs on a specific target node

        Returns:
            List of jobs in the format job_kind[job_id]
        """
        return self._client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_CLUSTER,
            res_model=List[str],
            json=JobQuery(kind=job_kind, target=target_id, active=True).as_dict(),
            params={QPARAM_WHAT: WHAT_ALL_RUNNING_STATUS},
        )

    def list_running_etls(self) -> List[ETLInfo]:
        """
        Lists all running ETLs.

        Note: Does not list ETLs that have been stopped or deleted.

        Returns:
            List[ETLInfo]: A list of details on running ETLs
        """
        return self._client.request_deserialize(
            HTTP_METHOD_GET, path=URL_PATH_ETL, res_model=List[ETLInfo]
        )

    def is_ready(self) -> bool:
        """
        Checks if cluster is ready or still setting up.

        Returns:
            bool: True if cluster is ready, or false if cluster is still setting up
        """
        # compare with AIS Go API (api/cluster.go) for additional supported options
        params = {QPARAM_PRIMARY_READY_REB: "true"}
        try:
            resp = self._client.request(
                HTTP_METHOD_GET,
                path=URL_PATH_HEALTH,
                endpoint=self.get_primary_url(),
                params=params,
            )
            return resp.ok
        except Exception as err:
            logger.debug(err)
            return False

    # pylint: disable=too-many-locals
    def get_performance(
        self,
        get_throughput: bool = True,
        get_latency: bool = True,
        get_counters: bool = True,
    ) -> ClusterPerformance:
        """
        Retrieves and calculates the performance metrics for each target node in the AIStore cluster.
        It compiles throughput, latency, and various operational counters from each target node,
        providing a comprehensive view of the cluster's overall performance

        Args:
            get_throughput (bool, optional): get cluster throughput
            get_latency (bool, optional): get cluster latency
            get_counters (bool, optional): get cluster counters

        Returns:
            ClusterPerformance: An object encapsulating the detailed performance metrics of the cluster,
                including throughput, latency, and counters for each node

        Raises:
            requests.RequestException: If there's an ambiguous exception while processing the request
            requests.ConnectionError: If there's a connection error with the cluster
            requests.ConnectionTimeout: If the connection to the cluster times out
            requests.ReadTimeout: If the timeout is reached while awaiting a response from the cluster
        """

        targets = self._get_targets()
        target_stats = {}
        params = {QPARAM_WHAT: WHAT_NODE_STATS_AND_STATUS}
        res_model = NodeStats
        for target_id in targets:
            headers = {HEADER_NODE_ID: target_id}
            try:
                res = self.client.request_deserialize(
                    HTTP_METHOD_GET,
                    path=f"{URL_PATH_REVERSE}/{URL_PATH_DAEMON}",
                    res_model=res_model,
                    headers=headers,
                    params=params,
                )
            except AISError as err:
                if "unrecognized what=node_status" in err.message:
                    params = {QPARAM_WHAT: WHAT_NODE_STATS_AND_STATUS_V322}
                    res_model = NodeStatsV322
                    res = self.client.request_deserialize(
                        HTTP_METHOD_GET,
                        path=f"{URL_PATH_REVERSE}/{URL_PATH_DAEMON}",
                        res_model=res_model,
                        headers=headers,
                        params=params,
                    )
                else:
                    raise err
            target_stats[target_id] = res
        throughputs = {}
        latencies = {}
        counters = {}
        for target_id, val in target_stats.items():
            if get_throughput:
                throughputs[target_id] = NodeThroughput(val.tracker)
            if get_latency:
                latencies[target_id] = NodeLatency(val.tracker)
            if get_counters:
                counters[target_id] = NodeCounter(val.tracker)

        return ClusterPerformance(
            throughput=throughputs, latency=latencies, counters=counters
        )

    def _get_smap(self):
        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_DAEMON,
            res_model=Smap,
            params={QPARAM_WHAT: WHAT_SMAP},
        )

    def _get_targets(self):
        tmap = self._get_smap().tmap
        return list(tmap.keys())

    def get_uuid(self) -> str:
        """
        Returns: UUID of AIStore Cluster
        """
        return self._get_smap().uuid



#
# Copyright (c) 2022-2023, NVIDIA CORPORATION. All rights reserved.
#
from io import BufferedWriter
from typing import Dict, NewType

import requests

from aistore.sdk.const import (
    DEFAULT_CHUNK_SIZE,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_HEAD,
    HTTP_METHOD_PUT,
    QPARAM_ARCHPATH,
    QPARAM_ARCHREGX,
    QPARAM_ARCHMODE,
    QPARAM_OBJ_APPEND,
    QPARAM_OBJ_APPEND_HANDLE,
    QPARAM_ETL_NAME,
    QPARAM_LATEST,
    QPARAM_NEW_CUSTOM,
    ACT_PROMOTE,
    HTTP_METHOD_PATCH,
    HTTP_METHOD_POST,
    URL_PATH_OBJECTS,
    HEADER_RANGE,
    HEADER_OBJECT_APPEND_HANDLE,
    ACT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_WORKERS,
    HEADER_OBJECT_BLOB_CHUNK_SIZE,
)
from aistore.sdk.object_reader import ObjectReader
from aistore.sdk.types import (
    ActionMsg,
    PromoteAPIArgs,
    BlobMsg,
    ArchiveSettings,
    BlobDownloadSettings,
)
from aistore.sdk.utils import read_file_bytes, validate_file
from aistore.sdk.object_props import ObjectProps

Header = NewType("Header", requests.structures.CaseInsensitiveDict)


# pylint: disable=consider-using-with,unused-variable
class Object:
    """
    A class representing an object of a bucket bound to a client.

    Args:
        bucket (Bucket): Bucket to which this object belongs
        name (str): name of object
        size (int, optional): size of object in bytes
        props (ObjectProps, optional): Properties of object
    """

    def __init__(self, bucket: "Bucket", name: str, props: ObjectProps = None):
        self._bucket = bucket
        self._client = bucket.client
        self._bck_name = bucket.name
        self._qparams = bucket.qparam
        self._name = name
        self._object_path = f"{URL_PATH_OBJECTS}/{ self._bck_name}/{ self.name }"
        self._props = props

    @property
    def bucket(self):
        """Bucket containing this object."""
        return self._bucket

    @property
    def name(self) -> str:
        """Name of this object."""
        return self._name

    @property
    def props(self) -> ObjectProps:
        """Properties of this object."""
        return self._props

    def head(self) -> Header:
        """
        Requests object properties and returns headers. Updates props.

        Returns:
            Response header with the object properties.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        headers = self._client.request(
            HTTP_METHOD_HEAD,
            path=self._object_path,
            params=self._qparams,
        ).headers
        self._props = ObjectProps(headers)
        return headers

    # pylint: disable=too-many-arguments
    def get(
        self,
        archive_settings: ArchiveSettings = None,
        blob_download_settings: BlobDownloadSettings = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl_name: str = None,
        writer: BufferedWriter = None,
        latest: bool = False,
        byte_range: str = None,
    ) -> ObjectReader:
        """
        Creates and returns an ObjectReader with access to object contents and optionally writes to a provided writer.

        Args:
            archive_settings (ArchiveSettings, optional): Settings for archive extraction
            blob_download_settings (BlobDownloadSettings, optional): Settings for using blob download
            chunk_size (int, optional): chunk_size to use while reading from stream
            etl_name (str, optional): Transforms an object based on ETL with etl_name
            writer (BufferedWriter, optional): User-provided writer for writing content output
                User is responsible for closing the writer
            latest (bool, optional): GET the latest object version from the associated remote bucket
            byte_range (str, optional): Specify a specific data segment of the object for transfer, including
                both the start and end of the range (e.g. "bytes=0-499" to request the first 500 bytes)

        Returns:
            An ObjectReader which can be iterated over to stream chunks of object content or used to read all content
            directly.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        params = self._qparams.copy()
        headers = {}
        if archive_settings:
            if archive_settings.mode:
                archive_settings.mode = archive_settings.mode.value
            params[QPARAM_ARCHPATH] = archive_settings.archpath
            params[QPARAM_ARCHREGX] = archive_settings.regex
            params[QPARAM_ARCHMODE] = archive_settings.mode

        if blob_download_settings:
            headers[HEADER_OBJECT_BLOB_DOWNLOAD] = "true"
            headers[HEADER_OBJECT_BLOB_CHUNK_SIZE] = blob_download_settings.chunk_size
            headers[HEADER_OBJECT_BLOB_WORKERS] = blob_download_settings.num_workers
        if etl_name:
            params[QPARAM_ETL_NAME] = etl_name
        if latest:
            params[QPARAM_LATEST] = "true"

        if byte_range and blob_download_settings:
            raise ValueError("Cannot use Byte Range with Blob Download")

        if byte_range:
            # For range formatting, see the spec:
            # https://www.rfc-editor.org/rfc/rfc7233#section-2.1
            headers = {HEADER_RANGE: byte_range}

        obj_reader = ObjectReader(
            client=self._client,
            path=self._object_path,
            params=params,
            headers=headers,
            chunk_size=chunk_size,
        )
        if writer:
            writer.writelines(obj_reader)
        return obj_reader

    def get_semantic_url(self):
        """
        Get the semantic URL to the object

        Returns:
            Semantic URL to get object
        """

        return f"{self.bucket.provider}://{self._bck_name}/{self._name}"

    def get_url(self, archpath: str = "", etl_name: str = None):
        """
        Get the full url to the object including base url and any query parameters

        Args:
            archpath (str, optional): If the object is an archive, use `archpath` to extract a single file
                from the archive
            etl_name (str, optional): Transforms an object based on ETL with etl_name

        Returns:
            Full URL to get object

        """
        params = self._qparams.copy()
        if archpath:
            params[QPARAM_ARCHPATH] = archpath
        if etl_name:
            params[QPARAM_ETL_NAME] = etl_name
        return self._client.get_full_url(self._object_path, params)

    def put_content(self, content: bytes) -> Header:
        """
        Puts bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to put as an object.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        self._put_data(self.name, content)

    def put_file(self, path: str = None):
        """
        Puts a local file as an object to a bucket in AIS storage.

        Args:
            path (str): Path to local file

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            ValueError: The path provided is not a valid file
        """
        validate_file(path)
        self._put_data(self.name, read_file_bytes(path))

    def _put_data(self, obj_name: str, data: bytes):
        url = f"{URL_PATH_OBJECTS}/{ self._bck_name }/{ obj_name }"
        self._client.request(
            HTTP_METHOD_PUT,
            path=url,
            params=self._qparams,
            data=data,
        )

    # pylint: disable=too-many-arguments
    def promote(
        self,
        path: str,
        target_id: str = "",
        recursive: bool = False,
        overwrite_dest: bool = False,
        delete_source: bool = False,
        src_not_file_share: bool = False,
    ) -> str:
        """
        Promotes a file or folder an AIS target can access to a bucket in AIS storage.
        These files can be either on the physical disk of an AIS target itself or on a network file system
        the cluster can access.
        See more info here: https://aiatscale.org/blog/2022/03/17/promote

        Args:
            path (str): Path to file or folder the AIS cluster can reach
            target_id (str, optional): Promote files from a specific target node
            recursive (bool, optional): Recursively promote objects from files in directories inside the path
            overwrite_dest (bool, optional): Overwrite objects already on AIS
            delete_source (bool, optional): Delete the source files when done promoting
            src_not_file_share (bool, optional): Optimize if the source is guaranteed to not be on a file share

        Returns:
            Job ID (as str) that can be used to check the status of the operation, or empty if job is done synchronously

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            AISError: Path does not exist on the AIS cluster storage
        """
        url = f"{URL_PATH_OBJECTS}/{ self._bck_name }"
        value = PromoteAPIArgs(
            source_path=path,
            object_name=self.name,
            target_id=target_id,
            recursive=recursive,
            overwrite_dest=overwrite_dest,
            delete_source=delete_source,
            src_not_file_share=src_not_file_share,
        ).as_dict()
        json_val = ActionMsg(action=ACT_PROMOTE, name=path, value=value).dict()

        return self._client.request(
            HTTP_METHOD_POST, path=url, params=self._qparams, json=json_val
        ).text

    def delete(self):
        """
        Delete an object from a bucket.

        Returns:
            None

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        self._client.request(
            HTTP_METHOD_DELETE,
            path=self._object_path,
            params=self._qparams,
        )

    def blob_download(
        self,
        chunk_size: int = None,
        num_workers: int = None,
        latest: bool = False,
    ) -> str:
        """
        A special facility to download very large remote objects a.k.a. BLOBs
        Returns job ID that for the blob download operation.

        Args:
            chunk_size (int): chunk size in bytes
            num_workers (int): number of concurrent blob-downloading workers (readers)
            latest (bool): GET the latest object version from the associated remote bucket

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
        """
        params = self._qparams.copy()
        value = BlobMsg(
            chunk_size=chunk_size,
            num_workers=num_workers,
            latest=latest,
        ).as_dict()
        json_val = ActionMsg(
            action=ACT_BLOB_DOWNLOAD, value=value, name=self.name
        ).dict()
        url = f"{URL_PATH_OBJECTS}/{ self._bck_name }"
        return self._client.request(
            HTTP_METHOD_POST, path=url, params=params, json=json_val
        ).text

    def append_content(
        self, content: bytes, handle: str = "", flush: bool = False
    ) -> str:
        """
        Append bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to append to the object.
            handle (str): Handle string to use for subsequent appends or flush (empty for the first append).
            flush (bool): Whether to flush and finalize the append operation, making the object accessible.

        Returns:
            handle (str): Handle string to pass for subsequent appends or flush.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """

        url = f"{URL_PATH_OBJECTS}/{ self._bck_name }/{ self.name }"
        params = self._qparams.copy()
        params[QPARAM_OBJ_APPEND] = "append" if not flush else "flush"
        params[QPARAM_OBJ_APPEND_HANDLE] = handle

        resp_headers = self._client.request(
            HTTP_METHOD_PUT,
            path=url,
            params=params,
            data=content,
        ).headers

        return resp_headers.get(HEADER_OBJECT_APPEND_HANDLE, "")

    def set_custom_props(
        self, custom_metadata: Dict[str, str], replace_existing: bool = False
    ):
        """
        Set custom properties for the object.

        Args:
            custom_metadata (Dict[str, str]): Custom metadata key-value pairs.
            replace_existing (bool, optional): Whether to replace existing metadata. Defaults to False.
        """
        params = self._qparams.copy()
        if replace_existing:
            params[QPARAM_NEW_CUSTOM] = "true"

        url = f"{URL_PATH_OBJECTS}/{self._bck_name}/{self.name}"

        json_val = ActionMsg(action="", value=custom_metadata).dict()

        self._client.request(HTTP_METHOD_PATCH, path=url, params=params, json=json_val)



#
# Copyright (c) 2022-2024, NVIDIA CORPORATION. All rights reserved.
#

import pydantic.tools

from aistore.sdk.utils import HttpError
from aistore.sdk.authn.errors import (
    AuthNError,
    ErrClusterNotFound,
    ErrClusterAlreadyRegistered,
    ErrRoleNotFound,
    ErrRoleAlreadyExists,
    ErrUserNotFound,
    ErrUserAlreadyExists,
    ErrUserInvalidCredentials,
)


def raise_authn_error(text: str):
    """
    Raises an AuthN-specific error based on the API response text.

    Args:
        text (str): The raw text of the API response containing error details.

    Raises:
        AuthNError: If the error doesn't match any specific conditions.
        ErrClusterNotFound: If the error message indicates a missing cluster.
        ErrRoleNotFound: If the error message indicates a missing role.
        ErrUserNotFound: If the error message indicates a missing user.
        ErrRoleAlreadyExists: If the error message indicates a role already exists.
        ErrUserAlreadyExists: If the error message indicates a user already exists.
        ErrClusterAlreadyRegistered: If the error message indicates the cluster is already registered.
        ErrUserInvalidCredentials: If the error message indicates invalid user credentials.
    """
    err = pydantic.tools.parse_raw_as(HttpError, text)
    if 400 <= err.status <= 500:
        if "does not exist" in err.message:
            if "cluster" in err.message:
                raise ErrClusterNotFound(err.status, err.message)
            if "role" in err.message:
                raise ErrRoleNotFound(err.status, err.message)
            if "user" in err.message:
                raise ErrUserNotFound(err.status, err.message)
        elif "already exists" in err.message:
            if "role" in err.message:
                raise ErrRoleAlreadyExists(err.status, err.message)
            if "user" in err.message:
                raise ErrUserAlreadyExists(err.status, err.message)
        elif "already registered" in err.message:
            raise ErrClusterAlreadyRegistered(err.status, err.message)
        elif "invalid credentials" in err.message:
            raise ErrUserInvalidCredentials(err.status, err.message)
    raise AuthNError(err.status, err.message)



from aistore.sdk.client import Client
from aistore.sdk.list_object_flag import ListObjectFlag
from aistore.sdk.bucket import Bucket
from aistore.sdk.namespace import Namespace
from aistore.sdk.authn.authn_client import AuthNClient



#
# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
#

from enum import Enum


class ArchiveMode(Enum):
    """
    Archive mode for getting files/objects from an archive in a bucket
    See `MatchMode` enum in the cmn/archive/read.go
    """

    REGEXP = "regexp"
    PREFIX = "prefix"
    SUFFIX = "suffix"
    SUBSTR = "substr"
    WDSKEY = "wdskey"



#
# Copyright (c) 2022-2024, NVIDIA CORPORATION. All rights reserved.
#
from pathlib import Path
from typing import Callable, Iterator, Tuple, Type, TypeVar
from urllib.parse import urlparse

import braceexpand
import humanize

from msgspec import msgpack
import pydantic.tools
import requests
from pydantic import BaseModel, parse_raw_as
from requests import Response

from aistore.sdk.const import UTF_ENCODING, HEADER_CONTENT_TYPE, MSGPACK_CONTENT_TYPE
from aistore.sdk.errors import (
    AISError,
    ErrBckNotFound,
    ErrRemoteBckNotFound,
    ErrBckAlreadyExists,
    ErrETLAlreadyExists,
)

T = TypeVar("T")


class HttpError(BaseModel):
    """
    Represents an error returned by the API.
    """

    status: int
    message: str = ""
    method: str = ""
    url_path: str = ""
    remote_addr: str = ""
    caller: str = ""
    node: str = ""


def raise_ais_error(text: str) -> None:
    """
    Raise an AIS error based on the response text.

    Args:
        text (str): The raw text of the API response containing error details.

    Raises:
        AISError: If the error doesn't match any specific conditions.
        ErrBckNotFound: If the error message indicates a missing bucket.
        ErrRemoteBckNotFound: If the error message indicates a missing remote bucket.
        ErrBckAlreadyExists: If the error message indicates a bucket already exists.
        ErrETLAlreadyExists: If the error message indicates an ETL already exists
    """
    err = pydantic.tools.parse_raw_as(HttpError, text)
    if 400 <= err.status < 500:
        if "does not exist" in err.message:
            if "cloud bucket" in err.message or "remote bucket" in err.message:
                raise ErrRemoteBckNotFound(err.status, err.message)
            if "bucket" in err.message:
                raise ErrBckNotFound(err.status, err.message)
        elif "already exists" in err.message:
            if "bucket" in err.message:
                raise ErrBckAlreadyExists(err.status, err.message)
            if "etl" in err.message:
                raise ErrETLAlreadyExists(err.status, err.message)

    raise AISError(err.status, err.message)


# pylint: disable=unused-variable
def handle_errors(
    resp: requests.Response, raise_error_fn: Callable[[str], None] = raise_ais_error
) -> None:
    """
    Error handling for requests made to the AIS Client

    Args:
        resp: requests.Response = Response received from the request
        raise_error_fn: Function that processes error text and raises appropriate exceptions.
    """
    error_text = resp.text
    if isinstance(resp.text, bytes):
        try:
            error_text = error_text.decode(UTF_ENCODING)
        except UnicodeDecodeError:
            error_text = error_text.decode("iso-8859-1")
    if error_text != "":
        raise_error_fn(error_text)
    resp.raise_for_status()


def probing_frequency(dur: int) -> float:
    """
    Given a timeout, return an interval to wait between retries

    Args:
        dur: Duration of timeout

    Returns:
        Frequency to probe
    """
    freq = min(dur / 8.0, 1.0)
    freq = max(dur / 64.0, freq)
    return max(freq, 0.1)


def read_file_bytes(filepath: str):
    """
    Given a filepath, read the content as bytes
    Args:
        filepath: Existing local filepath

    Returns: Raw bytes
    """
    with open(filepath, "rb") as reader:
        return reader.read()


def _check_path_exists(path: str) -> None:
    if not Path(path).exists():
        raise ValueError(f"Path: {path} does not exist")


def validate_file(path: str) -> None:
    """
    Validate that a file exists and is a file
    Args:
        path: Path to validate
    Raises:
        ValueError: If path does not exist or is not a file
    """
    _check_path_exists(path)
    path_obj = Path(path)
    if not path_obj.exists():
        raise ValueError(f"Path: {path} does not exist")
    if not path_obj.is_file():
        raise ValueError(f"Path: {path} is a directory, not a file")


def validate_directory(path: str) -> None:
    """
    Validate that a directory exists and is a directory
    Args:
        path: Path to validate
    Raises:
        ValueError: If path does not exist or is not a directory
    """
    _check_path_exists(path)
    if not Path(path).is_dir():
        raise ValueError(f"Path: {path} is a file, not a directory")


def get_file_size(file: Path) -> str:
    """
    Get the size of a file and return it in human-readable format
    Args:
        file: File to read

    Returns:
        Size of file as human-readable string

    """
    return (
        humanize.naturalsize(file.stat().st_size) if file.stat().st_size else "unknown"
    )


def expand_braces(template: str) -> Iterator[str]:
    """
    Given a string template, apply bash-style brace expansion to return a list of strings
    Args:
        template: Valid brace expansion input, e.g. prefix-{0..10..2}-gap-{11..15}-suffix

    Returns:
        Iterator of brace expansion output

    """
    # pylint: disable = fixme
    # TODO Build custom expansion to validate consistent with cmn/cos/template.go TemplateRange
    return braceexpand.braceexpand(template)


def decode_response(
    res_model: Type[T],
    resp: Response,
) -> T:
    """
    Parse response content from the cluster into a Python class,
     decoding with msgpack depending on content type in header

    Args:
        res_model (Type[T]): Resulting type to which the response should be deserialized
        resp (Response): Response from the AIS cluster

    """
    if resp.headers.get(HEADER_CONTENT_TYPE) == MSGPACK_CONTENT_TYPE:
        return msgpack.decode(resp.content, type=res_model)
    return parse_raw_as(res_model, resp.text)


def parse_url(url: str) -> Tuple[str, str, str]:
    """
    Parse AIS URLs for bucket and object names.

    Args:
        url (str): Complete URL of the object (e.g., "ais://bucket1/file.txt")

    Returns:
        Tuple[str, str, str]: Provider, bucket name, and object name
    """
    parsed_url = urlparse(url)
    path = parsed_url.path.lstrip("/")
    return parsed_url.scheme, parsed_url.netloc, path
